@@ -4,8 +4,8 @@
 from __future__ import print_function, unicode_literals
 
 import base64
-import copy
 import contextlib
+import copy
 import enum
 import functools
 import io
@@ -13,6 +13,8 @@ import json
 import logging
 import os
 import re
+import shutil
+import subprocess
 import threading
 import time
 from collections import defaultdict, namedtuple
@@ -24,11 +26,11 @@ import requests
 import retry
 import six
 from deprecated import deprecated
-from six.moves import urllib
 
 from . import requests_usbmux, xcui_element_types
 from ._proto import *
 from .exceptions import *
+from .usbmux import Usbmux
 from .utils import inject_call, limit_call_depth
 
 try:
@@ -38,6 +40,7 @@ except ImportError:
 
 try:
     import sys
+
     import logzero
     if not (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()):
         log_format = '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] %(message)s'
@@ -80,7 +83,7 @@ def convert(dictionary):
     return attrdict.AttrDict(dictionary)
 
     # Old implement
-    #return namedtuple('GenericDict', list(dictionary.keys()))(**dictionary)
+    # return namedtuple('GenericDict', list(dictionary.keys()))(**dictionary)
 
 
 def urljoin(*urls):
@@ -129,6 +132,7 @@ def _requests_session_pool_get(scheme, netloc):
 def _is_tmq_platform() -> bool:
     return os.getenv("TMQ") == "true"
 
+
 def _unsafe_httpdo(url, method='GET', data=None, timeout=None):
     """
     Do HTTP Request
@@ -168,7 +172,7 @@ def _unsafe_httpdo(url, method='GET', data=None, timeout=None):
             value = r.value.copy()
             value.pop("traceback", None)
 
-            for errCls in (WDAInvalidSessionIdError, WDAPossiblyCrashedError, WDAKeyboardNotPresentError, WDAUnknownError):
+            for errCls in (WDAInvalidSessionIdError, WDAPossiblyCrashedError, WDAKeyboardNotPresentError, WDAUnknownError, WDAStaleElementReferenceError):
                 if errCls.check(value):
                     raise errCls(status, value)
 
@@ -225,12 +229,30 @@ class Rect(list):
         return self.y + self.height
 
 
+def _start_wda_xctest(udid: str, wda_bundle_id=None) -> bool:
+    tins_path = shutil.which("tins")
+    if not tins_path:
+        return False
+    logger.info("WDA is not running, exec: tins xctest")
+    args = []
+    if udid:
+        args.extend(['-u', udid])
+    args.append('xctest')
+    if wda_bundle_id:
+        args.extend(['-B', wda_bundle_id])
+    p = subprocess.Popen([tins_path] + args)
+    time.sleep(3)
+    if p.poll() is not None:
+        logger.warning("tins xctest launch failed")
+        return False
+    return True
+
 class BaseClient(object):
     def __init__(self, url=None, _session_id=None):
         """
         Args:
             target (string): the device url
-        
+
         If target is empty, device url will set to env-var "DEVICE_URL" if defined else set to "http://localhost:8100"
         """
         if not url:
@@ -249,13 +271,20 @@ class BaseClient(object):
         if not _session_id:
             self._init_callback()
 
+        # u = urllib.parse.urlparse(self.__wda_url)
+        # if u.scheme == "usbmux" and not self.is_ready():
+        #     udid = u.netloc.split(":")[0]
+        #     if _start_wda_xctest(udid):
+        #         self.wait_ready()
+                # raise RuntimeError("xctest start failed")
+
     def _callback_fix_invalid_session_id(self, err: WDAError):
         """ 当遇到 invalid session id错误时，更新session id并重试 """
-        if isinstance(err, WDAInvalidSessionIdError): # and not self.__is_app:
+        if isinstance(err, WDAInvalidSessionIdError):  # and not self.__is_app:
             self.session_id = None
             return Callback.RET_RETRY
         if isinstance(err, WDAPossiblyCrashedError):
-            self.session_id = self.session().session_id # generate new sessionId
+            self.session_id = self.session().session_id  # generate new sessionId
             return Callback.RET_RETRY
         """ 等待设备恢复上线 """
 
@@ -306,7 +335,8 @@ class BaseClient(object):
         Args:
             filename: json log
         """
-        self.register_callback(Callback.HTTP_REQUEST_BEFORE, self._callback_json_report)
+        self.register_callback(
+            Callback.HTTP_REQUEST_BEFORE, self._callback_json_report)
 
     def is_ready(self) -> bool:
         try:
@@ -348,8 +378,11 @@ class BaseClient(object):
         # Can't use res.value['sessionId'] = ...
         return res.value
 
-    def register_callback(self, event_name: str, func: Callable):
-        self.__callbacks[event_name].append(func)
+    def register_callback(self, event_name: str, func: Callable, try_first: bool = False):
+        if try_first:
+            self.__callbacks[event_name].insert(0, func)
+        else:
+            self.__callbacks[event_name].append(func)
 
     def unregister_callback(self,
                             event_name: Optional[str] = None,
@@ -432,7 +465,7 @@ class BaseClient(object):
         return namedtuple("HTTPRequest", ['fetch', 'get', 'post'])(
             self._fetch,
             functools.partial(self._fetch, "GET"),
-            functools.partial(self._fetch, "POST")) # yapf: disable
+            functools.partial(self._fetch, "POST"))  # yapf: disable
 
     @property
     def _session_http(self):
@@ -440,7 +473,7 @@ class BaseClient(object):
             functools.partial(self._fetch, with_session=True),
             functools.partial(self._fetch, "GET", with_session=True),
             functools.partial(self._fetch, "POST", with_session=True),
-            functools.partial(self._fetch, "DELETE", with_session=True)) # yapf: disable
+            functools.partial(self._fetch, "DELETE", with_session=True))  # yapf: disable
 
     def home(self):
         """Press home button"""
@@ -498,10 +531,10 @@ class BaseClient(object):
         Args:
             png_filename(string): optional, save file name
             format(string): return format, "raw" or "pillow” (default)
-        
+
         Returns:
             PIL.Image or raw png data
-        
+
         Raises:
             WDARequestError
         """
@@ -661,17 +694,20 @@ class BaseClient(object):
         return self.session_id
 
     @cached_property
-    def scale(self):
+    def scale(self) -> int:
         """
         UIKit scale factor
-        
+
         Refs:
             https://developer.apple.com/library/archive/documentation/DeviceInformation/Reference/iOSDeviceCompatibility/Displays/Displays.html
         There is another way to get scale
-            self.http.get("/wda/screen").value returns {"statusBarSize": {'width': 320, 'height': 20}, 'scale': 2}
+            self._session_http.get("/wda/screen").value returns {"statusBarSize": {'width': 320, 'height': 20}, 'scale': 2}
         """
-        v = max(self.screenshot().size) / max(self.window_size())
-        return round(v)
+        try:
+            return self._session_http.get("/wda/screen").value['scale']
+        except (KeyError, WDARequestError):
+            v = max(self.screenshot().size) / max(self.window_size())
+            return round(v)
 
     @cached_property
     def bundle_id(self):
@@ -728,7 +764,7 @@ class BaseClient(object):
         """
         Args:
             callback (func): called when alert popup
-        
+
         Example of callback:
 
             def callback(session):
@@ -736,12 +772,12 @@ class BaseClient(object):
         """
         pass
 
-    #Not working
-    #def get_clipboard(self):
+    # Not working
+    # def get_clipboard(self):
     #    return self.http.post("/wda/getPasteboard").value
 
     # Not working
-    #def siri_activate(self, text):
+    # def siri_activate(self, text):
     #    self.http.post("/wda/siri/activate", {"text": text})
 
     def app_launch(self,
@@ -760,6 +796,10 @@ class BaseClient(object):
         assert isinstance(arguments, (tuple, list))
         assert isinstance(environment, dict)
 
+        # When device is locked, it is unable to launch
+        if self.locked():
+            self.unlock()
+
         return self._session_http.post(
             "/wda/apps/launch", {
                 "bundleId": bundle_id,
@@ -786,7 +826,7 @@ class BaseClient(object):
                 "value": 4,
                 "sessionId": "0363BDC5-4335-47ED-A54E-F7CCB65C6A65"
             }
-        
+
         value 1(not running) 2(running in background) 3(running in foreground)
         """
         return self._session_http.post("/wda/apps/state", {
@@ -812,7 +852,7 @@ class BaseClient(object):
 
         Returns:
             list of app
-        
+
         Return example:
             [{'pid': 52, 'bundleId': 'com.apple.springboard'}]
         """
@@ -824,10 +864,12 @@ class BaseClient(object):
         https://github.com/facebook/WebDriverAgent/blob/master/WebDriverAgentLib/Commands/FBSessionCommands.m#L43
         Args:
             url (str): url
-        
+
         Raises:
             WDARequestError
         """
+        if os.getenv("TMQ_ORIGIN") == "civita": # MDS platform
+            return self.http.post("/mds/openurl", {"url": url})
         return self._session_http.post('url', {'url': url})
 
     def deactivate(self, duration):
@@ -857,7 +899,7 @@ class BaseClient(object):
     def click(self, x, y, duration: Optional[float] = None):
         """
         Combine tap and tap_hold
-        
+
         Args:
             x, y: can be float(percent) or int
             duration (optional): tap_hold duration
@@ -969,8 +1011,22 @@ class BaseClient(object):
         size = self._unsafe_window_size()
         if min(size) > 0:
             return size
-        _ = self.orientation # after this operation, will safe to get window_size
-        return self._unsafe_window_size()
+
+        # get orientation, handle alert
+        _ = self.orientation  # after this operation, may safe to get window_size
+        if self.alert.exists:
+            self.alert.accept()
+            time.sleep(.1)
+
+        size = self._unsafe_window_size()
+        if min(size) > 0:
+            return size
+
+        logger.warning("unable to get window_size(), try to to create a new session")
+        with self.session("com.apple.Preferences") as app:
+            size = app._unsafe_window_size()
+            assert min(size) > 0, "unable to get window_size"
+            return size
 
     def _unsafe_window_size(self):
         """
@@ -1051,7 +1107,10 @@ class BaseClient(object):
 
 
 class Alert(object):
-    DEFAULT_ACCEPT_BUTTONS = []
+    DEFAULT_ACCEPT_BUTTONS = [
+        "使用App时允许", "无线局域网与蜂窝网络", "好", "稍后", "稍后提醒", "确定",
+        "允许", "以后", "打开", "录屏", "Allow", "OK", "YES", "Yes", "Later", "Close"
+    ]
 
     def __init__(self, client: BaseClient):
         self._c = client
@@ -1087,35 +1146,54 @@ class Alert(object):
     def buttons(self):
         return self.http.get('/wda/alert/buttons').value
 
-    def click(self, button_name: Union[str, list]):
+    def click(self, button_name: Optional[Union[str, list]] = None):
         """
         Args:
             - button_name: the name of the button
 
+        Returns:
+            button_name being clicked
+
         Raises:
             ValueError when button_name is not in avaliable button names
         """
         # Actually, It has no difference POST to accept or dismiss
         if isinstance(button_name, str):
-            return self.http.post('/alert/accept', data={"name": button_name})
+            self.http.post('/alert/accept', data={"name": button_name})
+            return button_name
+
         avaliable_names = self.buttons()
-        for bname in button_name:
+        buttons: list = button_name
+        for bname in buttons:
             if bname in avaliable_names:
-                return self.http.post('/alert/accept', data={"name": bname})
+                return self.click(bname)
         raise ValueError("Only these buttons can be clicked", avaliable_names)
 
+    def click_exists(self, buttons: Optional[Union[str, list]] = None):
+        """
+         Args:
+            - buttons: the name of the button of list of names
+
+        Returns:
+            button_name clicked or None
+        """
+        try:
+            return self.click(buttons)
+        except (ValueError, WDARequestError):
+            return None
+
     @contextlib.contextmanager
     def watch_and_click(self,
-                        buttons: Optional[list] = [
-                            "使用App时允许", "无线局域网与蜂窝网络", "好", "稍后", "稍后提醒", "确定",
-                            "允许", "以后", "打开", "录屏", "Allow", "OK", "YES", "Yes", "Later", "Close"
-                        ],
+                        buttons: Optional[list] = None,
                         interval: float = 2.0):
         """ watch and click button
         Args:
             buttons: buttons name which need to click
             interval: check interval
         """
+        if not buttons:
+            buttons = self.DEFAULT_ACCEPT_BUTTONS
+
         event = threading.Event()
 
         def _inner():
@@ -1194,14 +1272,14 @@ class Selector(object):
             xpath (str): xpath string, a little slow, but works fine
             timeout (float): maxium wait element time, default 10.0s
             index (int): index of founded elements
-        
+
         WDA use two key to find elements "using", "value"
         Examples:
         "using" can be on of 
             "partial link text", "link text"
             "name", "id", "accessibility id"
             "class name", "class chain", "xpath", "predicate string"
-        
+
         predicate string support many keys
             UID,
             accessibilityContainer,
@@ -1290,7 +1368,7 @@ class Selector(object):
         """
         Returns:
             element_ids (list(string)): example ['id1', 'id2']
-        
+
         HTTP example response:
         [
             {"ELEMENT": "E2FF5B2A-DBDF-4E67-9179-91609480D80A"},
@@ -1336,6 +1414,7 @@ class Selector(object):
             chain = chain + '[%d]' % self._index
         return chain
 
+    @retry.retry(WDAStaleElementReferenceError, tries=3, delay=.5, jitter=.2)
     def find_element_ids(self):
         elems = []
         if self._id:
@@ -1440,7 +1519,7 @@ class Selector(object):
 
         Args:
             timeout (float): timeout for wait
-        
+
         Returns:
             bool: if successfully clicked
         """
@@ -1450,14 +1529,14 @@ class Selector(object):
         e.click()
         return True
 
-    def wait(self, timeout=None, raise_error=True):
+    def wait(self, timeout=None, raise_error=False):
         """ alias of get
         Args:
             timeout (float): timeout seconds
-            raise_error (bool): default true, whether to raise error if element not found
-        
-        Raises:
-            WDAElementNotFoundError
+            raise_error (bool): default false, whether to raise error if element not found
+
+        Returns:
+            Element or None
         """
         return self.get(timeout=timeout, raise_error=raise_error)
 
@@ -1466,7 +1545,7 @@ class Selector(object):
         Args:
             timeout (float): default timeout
             raise_error (bool): return bool or raise error
-        
+
         Returns:
             bool: works when raise_error is False
 
@@ -1619,7 +1698,7 @@ class Element(object):
         Args:
             direction (str): one of "visible", "up", "down", "left", "right"
             distance (float): swipe distance, only works when direction is not "visible"
-               
+
         Raises:
             ValueError
 
@@ -1652,7 +1731,7 @@ class Element(object):
         Args:
             scale (float): scale must > 0
             velocity (float): velocity must be less than zero when scale is less than 1
-        
+
         Example:
             pinchIn  -> scale:0.5, velocity: -1
             pinchOut -> scale:2.0, velocity: 1
@@ -1675,31 +1754,24 @@ class Element(object):
 
 class USBClient(Client):
     """ connect device through unix:/var/run/usbmuxd """
+
     def __init__(self, udid: str = "", port: int = 8100, wda_bundle_id=None):
+        if not udid:
+            usbmux = Usbmux()
+            infos = usbmux.device_list()
+            if len(infos) == 0:
+                raise RuntimeError("no device connected")
+            elif len(infos) >= 2:
+                raise RuntimeError("more then one device connected")
+            udid = infos[0]['SerialNumber']
+
         super().__init__(url="usbmux://{}:{}".format(udid, port))
-        self.__udid = udid
-        self.__wda_bundle_id = wda_bundle_id
-
-        if not self.is_ready():
-            self._start_xctest()
-
-    def _start_xctest(self) -> bool:
-        import shutil
-        import subprocess
-        tins_path = shutil.which("tins")
-        if not tins_path:
-            return False
-
-        logger.info("WDA is not running, exec: tins xctest")
-        args = []
-        if self.__udid:
-            args.extend(['-u', self.__udid])
-        args.append('xctest')
-        if self.__wda_bundle_id:
-            args.extend(['-B', self.__wda_bundle_id])
-
-        p = subprocess.Popen([tins_path] + args)
-        time.sleep(3)
-        assert p.poll() is None
-        return self.wait_ready()
+        if self.is_ready():
+            return
 
+        if shutil.which("tins"):
+            ok = _start_wda_xctest(udid, wda_bundle_id)
+            if not ok:
+                raise RuntimeError("init wda xctest error")
+            if not self.wait_ready(timeout=20):
+                raise RuntimeError("wda xctest launched but check failed")
